import sys
sys.path.append("..") 

import markov362m as mk
import markov362m.examples as mke

m = mk.MarkovChain("My first chain")
m.add_state("A")
m.add_state("B")
m.add_state("Stop")
m.add_transition("A","B", probability = 0.5)
m.add_transition("A","A", probability = 0.5)
m.add_transition("B","A", probability = 0.5)
m.add_transition("B","Stop", probability = 0.5)
m.add_transition("Stop","Stop")
m.compute()

print(m.info())

mke.list_examples()



# -*- coding: utf-8 -*-
import networkx as nx
import numpy as np
import markov362m._classes
import markov362m._representation

DEFAULT_COLOR = "blue"

class MarkovChain(nx.DiGraph):
    """Basic Markov-chain class.
    
    MarkovChain is derived from the DiGraph class from networkx. It uses
    DiGraph to store states (nodes) and transitions (edges), as well as 
    some of their attributes. In addition to the data kept by DiGraph, it
    computes and stores information about various chain-theoretic
    properties. 

    Args:
        title (str): the title of the chain (optional, default "Untitled")
    """

    # basic methods
    def __init__(self, title="Untitled", **kwargs):

        # general
        nx.DiGraph.__init__(self, **kwargs)
        self.title = title

        # aliases for DiGraph members
        self.states = self.node 
        self.states = self.nodes 
        self.transitions = self.edges
        self.number_of_states = self.number_of_nodes # a function 
        self.number_of_transitions = self.number_of_edges # a function!


        # Markov-chain specific
        self.classes = []
        self.transient_states = []
        self.recurrent_states = []
        self.number_of_transient_states = None
        self.number_of_recurrent_states = None
        
        self.transient_state_indices = []
        self.recurrent_state_indices = []
        self.transient_class_indices = []
        self.recurrent_class_indices = []

        self.number_of_classes = None
        self.number_of_transient_classes = None
        self.number_of_recurrent_classes = None

        self.is_irreducible = None
        self.no_transient = None
        self.is_aperiodic = None
        
        self.P = None
        """Contains the transition matrix P of the chain"""
        self.Q = None
        """Contains the TxT matrix Q"""
        self.R = None
        """Contains the TxC matrix Q"""
        self.F = None
        """Contains the fundamental matrix F = (I-Q)^{-1}."""

        # index-to-state dictionaries
        self.I_to_S = {}
        """A dictionary that translates indices in P to state names"""
        self.S_to_I = {}
        """A dictionary that translates state names to indices in P"""
        self.T_to_S = {}
        """A dictionary that translates indices in T to state names"""
        self.C_to_S = {}
        """A dictionary that translates indices in C to state names"""
        self.S_to_T = {}
        """A dictionary that translates transient state names to indices in T"""
        self.S_to_C = {}
        """A dictionary that translates recurrent state names to indices in C"""

        # internal flags
        self._classes_computed = False
        self._P_matrix_computed = False
        self._QRF_matrix_computed = False

        # esthetics
        self.canvas = [150, 120]

    def add_state(self, state, label=None, color=DEFAULT_COLOR,
                  position=None, process=True, **kwargs):
        """Adds a state to a Markov chain. 

        Args:
            state (str): a string representing a state. 
        """
        # default/random label and position
        if not label:
            label = str(state)
        if not position:
            position = np.random.random(size=2)*self.canvas

        self.add_node(state, label=label, position=position, **kwargs)

        self._classes_processed = False
        self._matrix_computed = False
        self._QRF_matrix_computed = False

    def add_transition(self, state1, state2, probability = 1,
                       curve=0.0, label=None, **kwargs):
        """Adds a transition to a Markov Chain. 

        A transition with probability given by the argument probability ias
        added to the chain. If no probability is given, it will default to
        1. 

        args:
            state1 (str): the initial state 
            state2 (str): the terminal state 
            probability (float): the probability of the transition (defaults to 1).
        """
        assert probability > 0, \
            "A transition must have a positive probability."

        if not label:
            label = str(probability)

        if state1 not in self.states:
            self.add_state(state1)
        if state2 not in self.states:
            self.add_state(state2)

        self.add_edge(state1, state2,
                      curve=curve, weight=probability, label=label,
                      **kwargs)

        self._classes_processed = False
        self._matrix_computed = False
        self._QRF_matrix_computed = False

   
    def is_transient(self, state):
        """Checks whether the state is transient. The chain must be
        "computed" first.

        Args: 
            state (str): 

        Returns:
            bool: The return value. 
        """
            
        if not self._classes_computed:
            _compute_classes(self)
        return(self.nodes[state]["type"] == "T")

    def is_recurrent(self, state):
        """Checks whether the state is recurrent. The chain must be
        "computed" first.

        Args: 
            state (str): 

        Returns:
            bool: The return value. 
        """
        if not self._classes_computed:
            _compute_classes(self)
        return(self.nodes[state]["type"] == "R")

    def list_of_states(self):
        """Returns the list of all states in the chain."""
        return(self.states())
    
    def list_of_transitions(self):
        """Returns the list of all transitions (pairs or states) the chain."""
        return(self.states())

    def period(self, state):
        """Returns the period of the state. The chain must be "computed" first.

        Args: 
            state (str): 

        Returns:
            int: The period of "state".
        """
        if not self._classes_computed:
            _compute_classes(self)
        return(self.nodes[state]["period"])

    def class_number(self, state):
        """ Returns the index of the class of the state in the 
        list m.classes. The chain must be "computed" first. 

        Args:
            state (str):

        Returns:
            int: the index of the class containing state.
        """
        if not self._classes_computed:
            _compute_classes(self)
        return(self.nodes[state]["class"])

    def compute(self):
        """ Performs various computations on a MarkovChain class. It
        computes periods of all states, classifies them, finds, transient
        and recurrent states, and produces the matrices P, Q, R and F (if
        they exist).
        """
        markov362m._classes._compute_P_matrix(self)
        markov362m._classes._compute_classes(self)
        if not self.no_transient:
            markov362m._classes._compute_QRF_matrices(self)

    def _dump_states(self):
        out = ""
        state_records = self.states(data=True)
        for record in state_records:
            out += record[0]+"\n"
            for k, v in record[1].items():
                out += f"  {k}: {v}\n"
            out += "\n"
        return(out)

    def _dump_transitions(self):
        out = ""
        transition_records = self.transitions(data=True)
        for record in transition_records:
            out += record[0]+" -> "+record[1]+"\n"
            for k, v in record[2].items():
                out += f"  {k}: {v}\n"
            out += "\n"
        return(out)

    # re-setting attributes
    def set_probability(self, state1, state2, probability):
        """Sets the probability of an already existing transition. Should not
        be used unless you want to modify an existing transition. If
        probability=0, this removes the transition. 

        Args:
            state1 (str): the initial state
            state2 (str): the terminal state
            probability (float): the new probability of the transition.
        """
        if float(probability) == 0.0:
            self.remove_edge(state1, state2)
        else:
            self[state1][state2]["probability"] = probability
            self[state1][state2]["weight"] = probability

    def set_name(self, state, name):
        """Sets the name of an already existing state. Should not
        be used unless you want to modify an existing chain.

        Args:
            state1 (str): the initial state
        """
        nx.relabel_nodes(self, {state: name}, copy=False)

    def set_label(self, state, label):
        self.states[state]["label"] = label

    # setting aesthetic attributes
    def set_loop_direction(self, state, dir):
        #"""sets the loop direction for a state"""
        self.node[state]["loop"] = dir

    def set_color(self, state, color):
        #"""sets the color for a state """
        self.node[state]["color"] = color
    
    def set_canvas(self, l0, l1):
        #"""sets the canvas dimensions """
        self.canvas = [l0,l1]

    def set_curve(self, state1, state2, curve):
        #"""sets the curvature value for a transition"""
        self.edges[state1, state2]["curve"] = curve
  
    def set_layout(self, layout):
        # """Sets layout data from a dictionary

        # args:
            # layout (dict): a dictionary from states to arrays 
                # [x,y] of floats (coordinates)

        # """
        for state in layout:
            self.states[state]["position"]= layout[state]

    # dictionaries to vectors
    def dict_to_T_column(self, dct, value_for_omitted=0):
        """Returns a column vector of size T, whose values are taken from
        the dictionary dict from transient states to numbers. The value
        value_for_omitted (which defaults to 0) will be used for 
        states not included in dct.

        Args:
            dct (dict): a dictionary from strings to numbers
            value_for_omitted (int or float): used for missing states
        """

        return(np.array(
            [ [ dct.get(self.T_to_S[i], value_for_omitted) ] 
                for i in range(self.number_of_transient_states) ]
            ))
    
    def dict_to_T_row(self, dct, value_for_omitted=0):
        """Returns a row vector of size T, whose values are taken from
        the dictionary dict from transient states to numbers. The value
        value_for_omitted (which defaults to 0) will be used for 
        states not included in dct. 

        Args:
            dct (dict): a dictionary from strings to numbers
            value_for_omitted (int or float): used for missing states
        """
        return(np.array(
            [ dct.get(self.T_to_S[i], value_for_omitted) 
                for i in range(self.number_of_transient_states) ]
            ))
    

    def dict_to_P_row(self, dct, value_for_omitted=0):
        """Returns a row vector of size P (total number of states), whose
        values are taken from the dictionary dict from states to
        numbers. The value value_for_omitted (which defaults to 0) will be
        used for states not included in dct. 

        Args:
            dct (dict): a dictionary from strings to numbers
            value_for_omitted (int or float): used for missing states
        """
        return(np.array(
            [  dct.get(self.I_to_S[i], value_for_omitted)  
                for i in range(self.number_of_states()) ]
            ))
    
    def dict_to_P_column(self, dct, value_for_omitted=0):
        """Returns a row vector of size P (total number of states), whose
        values are taken from the dictionary dict from states to
        numbers. The value value_for_omitted (which defaults to 0) will be
        used for states not included in dct. 

        Args:
            dct (dict): a dictionary from strings to numbers
            value_for_omitted (int or float): used for missing states
        """
        return(np.array(
            [ [ dct.get(self.I_to_S[i], value_for_omitted) ] 
                for i in range(self.number_of_states()) ]
            ))
    def P_to_dict(self, vct):
        """Returns a dictionary from states to values, built from the 
        vector vct, according to the internal order of states of P. 

        Args:
            vct(list of numbers): a list of values, of size P.

        Returns:
            dict: a dictionary from states to values
        """
        v = vct.flatten()
        return( { self.I_to_S[i] : v[i] for i in range(len(row)) } )
    
    def T_to_dict(self, vct):
        """Returns a dictionary from transient states to values, built from the 
        vector vct, according to the internal order of states of T. 

        Args:
            vct(list of numbers): a list of values, of size P.

        Returns:
            dict: a dictionary from states to values
        """
        v = vct.flatten()
        return( { self.T_to_S[i] : v[i] for i in range(len(vct)) } )
    
    def C_to_dict(self, vct):
        """Returns a dictionary from recurrent states to values, built from the 
        vector vct, according to the internal order of states of C.

        Args:
            vct(list of numbers): a list of values, of size P.

        Returns:
            dict: a dictionary from states to values
        """
        v = vct.flatten()
        return( { self.C_to_S[i] : v[i] for i in range(len(vct)) } )
    
    # representation/information

    def info(self):
        """Gives basic info about chain (states and transitions only).

        Returns: 
            str: returns a formatted string. Looks best when printed (as in print(m.info()))
        """
        return(markov362m._representation.info(self))
    
    def info_classes(self):
        """Gives info about the class structure of the chain.

        Returns: 
            str: returns a formatted string. Looks best when printed (as in
            print(m.info_classes()))
        """
        return(markov362m._representation.info_classes(self))
   
    def info_P(self):
        """Info about the transition matrix P and the order of states.

        Returns: 
            str: returns a formatted string. Looks best when printed (as in
            print(m.info_P()))
        """
        return(markov362m._representation.info_P(self))
  
    def info_QRF(self):
        """Info about the matrices Q and R, as well as the 
        fundamental matrix F. 

        Returns: 
            str: returns a formatted string. Looks best when printed (as in
            print(m.info_QRF()))
        """
        return(markov362m._representation.info_QRF(self))

    def info_all(self):
        """All available info about the chain.
        
        Returns: 
            str: returns a formatted string. Looks best when printed (as in
            print(m.info_all()))
        """
        infos = [ self.info(),
                self.info_classes(),
                self.info_P(),
                self.info_QRF() ]
        return("\n".join(infos))




import os
import sys
import subprocess
import networkx as nx
from network2tikz import plot
import numpy as np
import ast
from wand.image import Image as WImage
from markov362m.base import *


def write_dot(m, file_name=None):
    """Outputs a Markov chain into a dot-format file. The file name
    defaults to the chain title + ".dot"."""
    if not file_name:
        file_name = m.title+".dot"

    nx.drawing.nx_agraph.write_dot(m.graph, file_name)


def read_dot(file_name):
    """Reads a Markov chain from a "dot" file"""

    g = nx.drawing.nx_agraph.read_dot(file_name)
    # "weight" from string into float
    weight_str = nx.get_edge_attributes(g, "weight")
    weight_float = {k: float(v) for k, v in weight_str.items()}
    nx.set_edge_attributes(g, weight_float, "weight")
    m = Markov()
    m.graph = g
    return(m)

def write_tex(m, file_name=None, vertex_size=11, margin=15, **kwargs):
    """Outputs a Markov chain object in a TikZ/LaTeX format. The file name
    defaults to the chain title + ".tex" """
    g = nx.DiGraph(m)

    # process directions for loops
    updown_offset = 0
    direction_map = {
        "u": 90+updown_offset,
        "d": -90+updown_offset,
        "l": 180+updown_offset,
        "r": 0+updown_offset,
        "ul": 135+updown_offset,
        "ur": 45+updown_offset,
        "dl": -135+updown_offset,
        "dr": -45+updown_offset
    }
    loop_dict_raw = nx.get_node_attributes(g, 'loop')
    loop_dict = {(k, k): direction_map[v]
                 for k, v in loop_dict_raw.items()}

    # process positions
    layout_dict_raw = nx.get_node_attributes(g, 'position')
    layout_dict = layout_dict_raw
    # layout_dict = {k: ast.literal_eval(v)
    # for k,v in layout_dict_raw.items()}

    # process curves
    curve_dict_raw = nx.get_edge_attributes(g, 'curve')
    curve_dict = {(k[0], k[1]): float(v) for k, v in curve_dict_raw.items()}

    # process edge labels
    label_dict_raw = nx.get_edge_attributes(g, 'label')
    label_dict = {(k[0], k[1]): v for k, v in label_dict_raw.items()}

    # process colors
    def color_map(n):
        if n == 'blue' or n == '0':
            return("blue!20!white")
        else:
            return("orange!50!white")

    color_dict_raw = nx.get_node_attributes(g, 'color')
    color_dict = {k: color_map(v) for k, v in color_dict_raw.items()}

    # process shapes
    def shape_map(n):
        if n == 0:
            return("circle")
        else:
            return("rectangle")

    shape_dict_raw = nx.get_node_attributes(g, 'shape')
    shape_dict = {k: shape_map(v) for k, v in shape_dict_raw.items()}

    # process canvas size

    # node styles
    visual_style = {}
    visual_style['vertex_size'] = vertex_size
    visual_style['vertex_shape'] = shape_dict
    visual_style['vertex_label'] = nx.get_node_attributes(g, 'label')
    visual_style['unit'] = 'mm'
    visual_style["layout"] = layout_dict
    visual_style["canvas"] = m.canvas
    visual_style['keep_aspect_ratio'] = True
    visual_style['vertex_color'] = color_dict
    visual_style['vertex_label_color'] = 'black'
    visual_style['vertex_label_size'] = 3.5
    visual_style['vertex_math_mode'] = False
    visual_style['margin'] = margin

    # edge styles
    visual_style['edge_color'] = 'black'
    visual_style['edge_curved'] = curve_dict
    visual_style['edge_label'] = label_dict
    visual_style['edge_label_size'] = 3.5
    visual_style['edge_arrow_size'] = .2
    visual_style['edge_arrow_width'] = .2
    visual_style['edge_loop_size'] = 20
    visual_style['edge_math_mode'] = True
    visual_style['edge_loop_position'] = loop_dict

    # Create a latex file and compile it
    # strip extension first
    if not file_name:
        file_name = m.title
    file_name = os.path.splitext(file_name)[0]

    plot(g, file_name+".tex", **visual_style)


def view(m, file_name=None, **kwargs):
    """Output the chain graph in a pdf format (via Latex/Tikz). The file
    name defaults to the chain title+".pdf" """
    if not file_name:
        file_name = m.title
    # strip extension
    file_name = os.path.splitext(file_name)[0]
    write_tex(m, file_name+".tex", **kwargs)
    texcommand = "pdflatex " + file_name+".tex"
    subprocess.run(texcommand, shell=True)
    for ext in ['aux', 'log']:
        subprocess.run("rm ./"+file_name+"."+ext, shell=True)
    subprocess.run("open "+file_name+".pdf", shell=True)


def display_pdf(m, file_name=None):
    """Used to display the (already built) pdf file in Jupyter"""
    if not file_name:
        file_name = m.title+".pdf"
    return(WImage(filename=file_name))



import numpy as np
np.set_printoptions(suppress=True, precision=3)

def info(m):
    out_title = f"Basic info for: \"{m.title}\"\n"
    out_underl = "-"*(len(out_title)-1)+"\n"
    transitions = " , ".join(
            [ "("+s1 + " -> "+ s2+ ", "
                    +str(np.float32(dct["weight"]))+")" for
            s1, s2, dct in m.transitions(data=True) ]
            )

    out = out_title + out_underl + \
    f"""\n - states ({m.number_of_states()}): {m.states}
 
 - transitions ({m.number_of_transitions()}): {transitions}
"""
    return(out)

def info_classes(m):
    assert m._classes_computed, "Compute classes first."

    out_title = f"Class info for: \"{m.title}\"\n"
    out_title += "-"*(len(out_title)-1)+"\n"
    if m.is_aperiodic:
        period_info = f"""\n - the chain is aperiodic (all states have period 1).\n"""
    else:
        period_info = f"""\n - the chain is not aperiodic. States with nontrivial periods: """
        dict_period = { state: m.states[state]["period"] \
                for state in m.states \
                if m.states[state]["period"]>1 }
        period_info += dict_period.__repr__() + "\n"

    if m.is_irreducible:
        reduc_info = "\n - the chain is irreducible (single, recurrent class).\n\n"
    else:
        reduc_info = f"\n - the chain is not irreducible. It has {m.number_of_classes} classes. \n"

    out_classes = out_title + period_info 
    out_classes += reduc_info
    transient_classes = [m.classes[i] for i in m.transient_class_indices]
    recurrent_classes = [m.classes[i] for i in range(m.number_of_classes)\
            if i not in m.transient_class_indices]

    out_classes +=f"""\n - classes ({m.number_of_classes}): {m.classes}

 - transient classes ({m.number_of_transient_classes}, T = {m.number_of_transient_states}): {transient_classes}

 - recurrent classes ({m.number_of_recurrent_classes}, C = {m.number_of_recurrent_states}): {recurrent_classes}
"""
    return(out_classes)

def info_P(m):
    assert m._P_matrix_computed, "Compute the matrix P first."
    out_title = f"P-matrix info for: \"{m.title}\"\n"
    out_title += "-"*(len(out_title)-1)+"\n"
    out = out_title+f""" - transition matrix: 
{m.P}
"""
    out += "\n - order of states:"+ m.I_to_S.__repr__()+"\n"
    return(out)

def info_QRF(m):
    out_title = f"QRF info for: \"{m.title}\"\n"
    out_title += "-"*(len(out_title)-1)+"\n"
    if m.no_transient:
        out = "\n - no transient states. Q, R and F are not defined."
    else:
        assert m._QRF_matrix_computed, "Compute QRF first."
        out= f"""\n - matrix Q (T x T): 
   {m.Q}

 - matrix R (T x C): 
   {m.R}

 - fundamental matrix F=(I-Q)^{-1} (T x T): 
   {m.F}
"""

        out += "\n - order of T-states: "+ m.T_to_S.__repr__()+"\n"
        out += "\n - order of C-states: "+ m.C_to_S.__repr__()
    return(out_title+out)
  



import sys
sys.path.append("..") 

import markov362m as mk
import markov362m.examples as mke
import markov362m.io as mio

m2 = mke.professor()


m = mke.facility(p=0.3)
m.set_probability("2-0-0-0","2-0-0-0", 1)
m.set_probability("2-0-0-0","0-0-1-1",0)
m.set_probability("2-0-0-0","1-0-1-0",0)
m.compute()
mio.view(m)




from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='markov362m',  # Required
    version='0.0.5',  # Required

    # This is a one-line description or tagline of what your project does. This
    # corresponds to the "Summary" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#summary
    description='A Markov-chain class for M362M', # Optional
    long_description=long_description,  # Optional
    url='https://github.com/gordanz/Markov362M',  # Optional
    author='Gordan Zitkovic',
    author_email='gordanz@math.utexas.edu',  # Optional
    #keywords='sample setuptools development',  # Optional
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),  # Required
    python_requires='>=3.5, <4',
    install_requires=['numpy','scipy','networkx>=2.3'],  # Optional
    project_urls={  # Optional
        'Source': 'https://github.com/gordanz/MarkovM362M/'
    },
)



import math
import sys, inspect
import networkx as nx
from networkx.drawing.nx_agraph import graphviz_layout 
from markov362m.base import MarkovChain

def _good_function(mod, func_name, func):
    return (  inspect.isfunction(func) and inspect.getmodule(func) == mod and
not func_name[0] == '_' and not func_name == "list_examples" )

def list_examples():
    """Prints out the names of all  currently implemented examples."""

    this_module = sys.modules[__name__] 
    dct = inspect.getmembers(this_module)
    good = [ k for k,v in dct if _good_function(this_module,k,v) ]
    print(sorted(good))

def _rst(alpha, sx, sy, dx, dy, x,y):
    """Helper for rotate_scale_translate"""

    PI_OVER_180 = 0.0174533
    cosa = math.cos(alpha*PI_OVER_180)
    sina = math.sin(alpha*PI_OVER_180)
    xp = sx*(cosa * x + sina * y)+dx
    yp = sy*(sina * x - cosa * y)+dy
    return([xp,yp])

def _rotate_scale_translate(point_dict, angle = 0, stretch_x = 1, 
        stretch_y = 1,  move_x = 0, move_y = 0):
    """Applies an affine transformation to a layout dictionary 
    for a Markov chain

    args:
        point_dict (dict): dictionary of states to positions [x,y]
        angle (float, optional) : angle of rotation
        stretch_x(float, optional) : stretch factor in x, defaults to 1
        stretch_y(float, optional) : stretch factor in y, defaults to 1
        move_x(float, optional) : translation in x, defaults to 1
        move_y(float, optional) : translation in y, defaults to 0

    Returns:
        a layout dictionary (dict of positions [x,y,z,...] with z, ...
        discarded)

    """
    return( { k:_rst(angle, stretch_x, stretch_y, move_x, move_y, v[0], v[1]) 
        for k,v in point_dict.items()} )

def attached_cycles(n1 = 4, n2 = 6, pA = 0.5, tail = False):
    """A Markov chain with two attached cycles and a tail.

    Args:
        n1 (int) : size of the first cycle
        n2 (int) : size of the second cycle
        pA (float) : probability of choosing the first cycle
        tail (bool) : whether to add a tail to the chain

    Returns: 
        a MarkovChain object

    """
    assert n1 > 1 and n2 > 1, \
            "Cycles need to be at least 2 states in length"
    
    m=MarkovChain(title="Attached_Cycles")

    # nodes (must make them explicitly blue if you use graphviz)
    m.add_state("AB1")

    for i in range(2,n1+1):
        m.add_state("A"+str(i))
    
    for i in range(2,n2+1):
        m.add_state("B"+str(i))

    m.add_transition("AB1","A2", probability = pA)
    m.add_transition("AB1","B2", probability = pA)
    m.add_transition("A"+str(n1),"AB1")
    m.add_transition("B"+str(n2),"AB1")
    if n1 > 2:
        for i in range(2,n1):
            m.add_transition("A"+str(i), "A"+str(i+1))
    if n2 > 2:
        for i in range(2,n2):
            m.add_transition("B"+str(i), "B"+str(i+1))

    # adding a tail
    if tail:
        m.add_state("C2")
        m.add_transition("AB1","C2")
        m.add_transition("C2","C2")
        m.set_probability("AB1","A2",pA/2)
        m.set_probability("AB1","C2",pA/2)

    # Layout
    ###############
    # pre_layout = graphviz_layout(m.graph, prog="fdp", 
            # args = "-Gmodel=subset -Gnodesep=1.5") 
    pre_layout = nx.kamada_kawai_layout(m, weight=1)
    layout = _rotate_scale_translate(pre_layout, angle = -90, 
            stretch_x = 2, stretch_y = 2)
    for s,pos in layout.items():
        m.states[s]["position"]= pos
    m.set_canvas(180,150)
    m.compute()
    return(m)

def gambler(a=5, p=0.25):
    """The chain for "Gambler's ruin" problem.

    Args: 
        p (float): the probability of winning in a single round
        a (int): the goal amount

    Returns: 
        a MarkovChain object.
    """
    
    assert a>2, "a must be at least 3"
    assert 0<p and p<1, "p must be in (0,1)"

    m=MarkovChain(title="Gambler")

    # nodes (must make them explicitly blue if you use graphviz)
    for i in range(a+1):
        m.add_state(str(i), color="blue", position = [i,0]);
        
    m.set_color(str(0),"orange")
    m.set_color(str(a),"orange")

    # transitions
    for i in range(1,a):
        m.add_transition(str(i), str(i-1), 
                probability = 1-p,
                label = "1-p",curve = 0.2)
        m.add_transition(str(i), str(i+1), 
                probability = p,
                label = "p", curve = 0.2)
    m.set_curve(str(1), str(0),0)
    m.set_curve(str(a-1), str(a),0)

    m.add_transition(str(0), str(0))
    m.add_transition(str(a), str(a))
    m.set_loop_direction(str(0),"u")
    m.set_loop_direction(str(a),"u")

    # Layout
    ###############

    m.set_canvas(180,150)
    m.compute()
    return(m)


def mc8_1():
    # """The chain for problem MC8_1 
    
    # Returns: a MarkovChain object
    # """
    
    m=MarkovChain(title="MC8_1")

    # nodes (must make them explicitly blue if you use graphviz)
    for i in range(8):
        m.add_state(str(i+1), color="blue");

    # From 1
    m.add_transition("1","1",probability = 0.5, label="1/2")
    m.add_transition("1","2",probability = 0.5, label="1/2")

    # From 2
    m.add_transition("2","3",probability = 0.5, label="1/2")
    m.add_transition("2","7",probability = 0.5, label="1/2")

    # From 3
    m.add_transition("3","4",curve = 0.2)

    # From 4
    m.add_transition("4","3",probability = 0.5, label="1/2",curve=0.2);
    m.add_transition("4","5",probability = 0.5, label="1/2")

    # From 5
    m.add_transition("5","6")

    # From 6
    m.add_transition("6","3",probability = 0.5, label="1/2")
    m.add_transition("6","6",probability = 0.5, label="1/2")

    # From 7
    m.add_transition("7","8",curve=0.2);

    # From 5
    m.add_transition("8","7",probability = 0.75, label="3/4", curve=0.2);
    m.add_transition("8","8",probability = 0.25, label="1/4")

    # loops
    m.set_loop_direction("1","ur")
    m.set_loop_direction("8","ur")
    m.set_loop_direction("6","ur")

    # Layout
    ###############
    pre_layout = graphviz_layout(m, prog="fdp", 
            args = "-Gmodel=subset -Gnodesep=1.5") 
    layout = _rotate_scale_translate(pre_layout, angle = -90, 
            stretch_x = 2, stretch_y = 2)
    m.set_layout(layout)
    m.set_canvas(180,150)
    m.compute()
    return(m)

def facility(p=0.4):
    """The chain for problem about the airline reservation system and its
    computers which get repaired
    
    Args: 
        p (float): the probability of a computer breakdown

    Returns: 
        a MarkovChain object
    """

    m = MarkovChain(title = "MC14-1")

    # states
    m.add_state("0-0-1-1", shape = 1);
    m.add_state("2-0-0-0", shape = 1);
    m.add_state("1-1-0-0", shape = 1);
    m.add_state("0-1-0-1", shape = 1);
    m.add_state("1-0-1-0", shape = 1);
    
    # edges
    m.add_transition("0-0-1-1","0-1-0-1",
            label="1", probability = 1)
    m.add_transition("2-0-0-0","0-0-1-1",
            label="p^2", probability = p**2 )
    m.add_transition("2-0-0-0","2-0-0-0",
            label="(1-p)^2", probability = (1-p)**2)
    m.add_transition("2-0-0-0","1-0-1-0",
            label="2p(1-p)", probability = 2* p*(1-p) )
    m.add_transition("1-1-0-0","2-0-0-0",
            label="1-p", probability = (1-p) )
    m.add_transition("1-1-0-0","1-0-1-0",
            label="p",curve=0.2, probability = p)
    m.add_transition("1-0-1-0","1-1-0-0",
            label="1-p",curve=0.2, probability = (1-p) )
    m.add_transition("1-0-1-0","0-1-0-1",
            label="p",curve=0.2, probability = p )
    m.add_transition("0-1-0-1","1-0-1-0",
            label="1",curve=0.2, probability = 1 )
    
    # loops
    m.set_loop_direction("2-0-0-0","d")
    
    # Layout
    pre_layout =  nx.kamada_kawai_layout(m, weight = 1)
    layout = _rotate_scale_translate(pre_layout, angle = -80)
    m.set_layout(layout)
    m.compute()
    return(m)

def mc20_1():
    # """The chain for problem MC20_1."""
    
    m = MarkovChain(title="MC20_1")
    g = m

    m.set_canvas(140,130)

    # nodes
    for i in range(7):
        m.add_state(str(i))

    the_transitions = {
            (1,2,1,"1"),
            (2,3,1,"1"),
            (3,4,1,"1"),
            (4,5,1,"1"),
            (5,1,1,"1"),
            (0,1,0.5,"1/2"),
            (0,6,0.5,"1/2"),
            (6,6,1,"1")
            }

    for (i1,i2,p,l) in the_transitions:
        m.add_transition(str(i1), str(i2), probability = p, label = l)

    # pre_layout =  nx.kamada_kawai_layout(g, weight = 1)
    pre_layout = graphviz_layout(g, prog="neato", args = "-Gnodesep=1.5") 
    layout = _rotate_scale_translate(pre_layout, angle = -90, stretch_x = 20, 
            stretch_y = 20)
    m.set_layout(layout)
    # m.compute()
    return(m)


def mc21_1():
    # """The chain for problem MC21_1
    
    # Returns: a MarkovChain object
    # """

    m = Markov(title="MC21_1")
    g = m
   
    # states
    for state in range(1,3+1):
        m.add_state(str(state))
   
    # transitions
    m.add_transition( "1","2", probability = 1, label = "1", curve = 0.1)
    m.add_transition( "2","1", probability = 0.25, label = "1/4", curve = 0.1)
    m.add_transition( "2","2", probability = 0.50, label = "1/2")
    m.add_transition( "2","3", probability = 0.25, label = "1/4")
    m.add_transition( "3","3", probability = 1, label = "1", color="orange")
   
    # loop directions
    m.set_loop_direction("2","d")
    m.set_loop_direction("3","d")
    
    # the layout
    pre_layout = graphviz_layout(g, prog="dot", args = "-GNodesep=1")
    the_layout = _rotate_scale_translate(pre_layout, angle = 90)
    m.set_layout(the_layout)
    m.compute()
    return(m)

def professor(p_morning = 0.05, p_afternoon = 0.20):
    """The Markov chain with the professor and umbrellas.

    Args:
        p_morning (float): probability of raining in the morning. Defaults
            to 0.05.
        p_afternoon (float): probability of raining in the afternoon.
            Defaults to 0.2.

    Returns: 
        a MarkovChain object

    """
    # constants
    q_morning = 1-p_morning
    q_afternoon = 1-p_afternoon
    p_morning_label = "p_m"
    q_morning_label = "q_m"
    p_afternoon_label = "p_a"
    q_afternoon_label = "q_a"
    
    
    m = MarkovChain(title = "professor")
    g = m
    
    for p in ["H", "O"]:
        for u in range(6):
            m.add_state( p+str(u))
    
    # absorbing states get a different color and a loop direction
    m.set_color("H5","orange")
    m.set_color("O5","orange")
    
    m.set_loop_direction("H5","d")
    m.set_loop_direction("O5","u")
    
    # add edges to g
    for i in range(5):
        init = "H" + str(i)
        term = "O" + str(4 - i)
        m.add_transition(init, term, 
                probability=q_morning, 
                label=q_morning_label, 
                curve = 0.2)
    
        if i>0:
            init = "H" + str(i)
            term = "O" + str(5 - i)
            m.add_transition(init, term, 
                    probability=p_morning, 
                    label=p_morning_label, 
                    curve = 0.2)
    
        init = "O" + str(i)
        term = "H" + str(4 - i)
        m.add_transition(init, term, 
                probability=q_afternoon, 
                label=q_afternoon_label, 
                curve = 0.2)
    
        if i>0:
            init = "O" + str(i)
            term = "H" + str(5 - i)
            m.add_transition(init, term, 
                    probability=p_afternoon, 
                    label=p_afternoon_label, 
                    curve = 0.2)
    
    m.add_transition("H0", "H5", 
            probability=p_morning, 
            label=p_morning_label)
    m.add_transition("O0", "O5", 
            probability=p_afternoon, 
            label=p_afternoon_label)
    m.add_transition("H5", "H5")
    m.add_transition("O5", "O5")
    
    # make arrow between O2 and H2 curve a bit more
    m.set_curve("O2","H2",0.5)
    m.set_curve("H2","O2",0.5)
    
    # Layout
    
    manual_layout = {
            'H5': [0.2,0], 'H0': [1,0], 'O4': [2,0], 
            'H1': [3,0], 'O3': [4,0], 'H2': [5,0],
            'O2': [5,1], 'H3': [4,1], 'O1': [3,1], 
            'H4': [2,1], 'O0': [1,1], 'O5': [0.2,1]
            }

    m.set_canvas(220,100)
    the_layout = _rotate_scale_translate(manual_layout, 
            stretch_x =  10, stretch_y = 10, 
            move_x = 20, move_y = 20)
    m.set_layout(the_layout)
    
    # last minute tweaks
    m.set_name("H5","Hw")
    m.set_name("O5","Ow")
    m.set_label("Hw","Hw")
    m.set_label("Ow","Ow")

    m.compute()
    return(m)

def tennis(p = 0.4):
    """The tennis chain.

    Args:
        p (float): the probability that S wins in a single rally

    Returns: 
        a MarkovChain object

    """
    m = MarkovChain(title = "tennis")
    g = m

    points = ["0","15","30","40","A"]
    
    def label(l):
        [i,j] = l
        if i<5 and j<5:
            return(""+points[i]+"-"+points[j]+"")
        elif i==5:
            return("$S_{win}$")
        else:
            return("$R_{win}$")
    
    def name(l):
        [i,j] = l
        if i<5 and j<5:
            return(""+points[i]+"-"+points[j]+"")
        elif i==5:
            return("S")
        else:
            return("R")
    
    
    # BUILDING THE GRAPH
    nodes_game = [ [i,j] for i in range(4) for j in range(4) ]
    nodes_other = [ [3,4], [4,3], [5,0], [0,5] ]
    nodes = nodes_game+nodes_other
    
    for nd in nodes:
        clr = "blue"
        if nd == [5,0] or nd == [0,5]:
            clr = "orange"
        m.add_state(name(nd),label=label(nd),color=clr)
   
    m.set_loop_direction(name([5,0]),"d")
    m.set_loop_direction(name([0,5]),"u")

    p_label = 'p'
    q_label = 'q'
    q=1-p
    
    for i in range(4):
        for j in range(4):
            if i<3:
                m.add_transition(name((i,j)), name((i+1,j)),
                        label=p_label, probability = p)
            elif j<3:
                m.add_transition(name((i,j)), name((5,0)),
                        label=p_label, probability = p)
            if j<3: 
                m.add_transition(name((i,j)), name((i,j+1)),
                        label=q_label, probability = q)
            elif i<3:
                m.add_transition(name((i,j)), name((0,5)),
                        label=q_label, probability = q)

    # advantages back and forth
    m.add_transition(name((3,3)),name((3,4)),
            label=q_label, curve=0.2, probability=q)
    m.add_transition(name((3,3)),name((4,3)),
            label=p_label, curve=0.2, probability=p)
    m.add_transition(name((4,3)),name((3,3)),
            label=q_label, curve=0.2, probability=q)
    m.add_transition(name((3,4)),name((3,3)),
            label=p_label, curve=0.2, probability=p)
    
    # wins
    m.add_transition(name((3,4)),name((0,5)),
            label=q_label,  probability=q)
    m.add_transition(name((4,3)),name((5,0)),
            label=p_label,  probability=p)
    m.add_transition(name((5,0)),name((5,0)))
    m.add_transition(name((0,5)),name((0,5)))
    
    
    # Layout
    m.set_canvas(160,110)
    pre_layout =  nx.kamada_kawai_layout(g, weight = 1)
    the_layout = _rotate_scale_translate(pre_layout, 
            angle = -104, stretch_x = 20, stretch_y = 20, 
            move_x = 30, move_y = 30)
    m.set_layout(the_layout)
    m.compute()
    return(m)

def patterns_HTH():
    """A chain used to analyze coin-tossing patterns"""

    m=MarkovChain(title = "Patterns")
    states = ["0","H","HT","Win"]
    for i,s in enumerate(states):
        m.add_state(s, position = [float(i),0])
    m.set_color("Win","orange")
    m.add_transition("0","0", probability = 0.5, label = "1/2")
    m.add_transition("0","H",probability = 0.5, label = "1/2")
    m.add_transition("H","H",probability = 0.5, label = "1/2")
    m.add_transition("H","HT",probability = 0.5, label = "1/2")
    m.add_transition("HT","0",probability = 0.5, label = "1/2", curve = -0.3)
    m.add_transition("HT","Win",probability = 0.5, label = "1/2")
    m.add_transition("Win","Win")
    m.set_loop_direction("0","u")
    m.set_loop_direction("H","u")
    m.set_loop_direction("Win","u")

    m.set_canvas(180,50)
    m.compute()
    return(m)

def patterns_HHH():
    """A chain used to analyze coin-tossing patterns"""

    m=MarkovChain(title = "Patterns")
    states = ["0","H","HH","Win"]
    for i,s in enumerate(states):
        m.add_state(s, position = [float(i),0])
    m.set_color("Win","orange")
    m.add_transition("0","0", probability = 0.5, label = "1/2")
    m.add_transition("0","H",probability = 0.5, label = "1/2")
    m.add_transition("H","0",probability = 0.5, label = "1/2", curve = 0.3)
    m.add_transition("H","HH",probability = 0.5, label = "1/2")
    m.add_transition("HH","0",probability = 0.5, label = "1/2", curve = -0.3)
    m.add_transition("HH","Win",probability = 0.5, label = "1/2")
    m.add_transition("Win","Win")

    m.set_loop_direction("0","u")
    m.set_loop_direction("H","u")
    m.set_loop_direction("Win","u")

    m.set_canvas(180,50)
    m.compute()
    return(m)
